<?php declare(strict_types=1);


namespace Drupal\entity_usage_updater;

use Drupal\Component\Plugin\Exception\PluginException;
use Drupal\Component\Plugin\PluginManagerInterface;
use Drupal\Core\Batch\BatchBuilder;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Entity\RevisionableInterface;
use Drupal\Core\Entity\RevisionableStorageInterface;
use Drupal\Core\Entity\RevisionLogInterface;
use Drupal\Core\Entity\TranslatableInterface;
use Drupal\Core\Entity\TranslatableRevisionableInterface;
use Drupal\Core\Entity\TranslatableRevisionableStorageInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Site\Settings;
use Drupal\entity_usage\EntityUsageInterface;

/**
 * Updates references to entities tracked with Entity Usage.
 *
 * @todo Add lock.
 * @todo Support optional entity validation prior to saving.
 * @todo Check that ERR fields aren't translatable (which we assume).
 * @todo check that content moderation isn't enabled for entities being updated?
 * @todo Sanitize exceptions.
 * @todo Improve details in error logging.
 */
class EntityUsageUpdater implements EntityUsageUpdaterInterface {

  /**
   * {@inheritdoc}
   */
  public function update(array $mapping): void {
    $batch_builder = new BatchBuilder();
    $batch_builder
      ->setFinishCallback([static::class, 'finish'])
      ->setTitle(t('Entity reference update'))
      ->setInitMessage(t('Updating entity references.'))
      ->setErrorMessage(t('There was an error updating entity references.'));

    foreach ($mapping as $old_entity_type_id => $entity_updates) {
      foreach ($entity_updates as $old_id => [$new_entity_type_id, $new_id]) {
        $callback = [static::class, 'createBatchOperationsFromUsages'];
        $arguments = [
          $old_entity_type_id,
          $old_id,
          $new_entity_type_id,
          $new_id,
        ];
        $batch_builder->addOperation($callback, $arguments);
      }
    }

    batch_set($batch_builder->toArray());
  }

  /**
   * Finds usages of a particular entity and adds batch update operations.
   *
   * It exists as a standalone batch operation in case the usages of a single
   * entity are particularly large.
   *
   * @param string $old_entity_type_id
   *   The entity type of the entity we're searching for.
   * @param string|int $old_id
   *   The entity ID of the entity we're searching for.
   * @param string $new_entity_type_id
   *   The entity type of the entity to replace it with.
   * @param string|int $new_id
   *   The entity ID of the entity to replace it with.
   * @param array &$context
   *   The batch context.
   */
  public static function createBatchOperationsFromUsages(string $old_entity_type_id, $old_id, string $new_entity_type_id, $new_id, array &$context): void {
    $entity_usage = \Drupal::service('entity_usage.usage');
    assert($entity_usage instanceof EntityUsageInterface);
    $entity = \Drupal::entityTypeManager()->getStorage($old_entity_type_id)->load($old_id);
    if (!$entity) {
      $translation_context = [
        '@entity_type_id' => $old_entity_type_id,
        '@id' => $old_id,
      ];
      $context['results']['errors'][] = t("Couldn't load @entity_type @id", $translation_context);
      return;
    }

    $usages = $entity_usage->listSources($entity);
    if (empty($usages)) {
      return;
    }

    $batch_builder = new BatchBuilder();
    $batch_builder
      ->setFinishCallback([static::class, 'finish'])
      ->setTitle(t('Entity reference update'))
      ->setInitMessage(t('Updating entity references.'))
      ->setErrorMessage(t('There was an error updating entity references.'));

    foreach ($usages as $entity_type_id => $entity_type_usages) {
      // For revisionable entity types, we're only interested in references that
      // are in the active revision.
      if (\Drupal::entityTypeManager()->getDefinition($entity_type_id)->isRevisionable()) {
        foreach ($entity_type_usages as $entity_id => $entity_usages) {
          $is_active = fn(array $usage): bool => static::getActiveRevisionId($entity_type_id, $entity_id, $usage['source_langcode']) === (int) $usage['source_vid'];
          $usages[$entity_type_id][$entity_id] = array_values(array_filter($entity_usages, $is_active));
        }
      }
      // Remove any empty arrays (ie entities that only had references in old
      // revisions).
      $usages[$entity_type_id] = array_filter($usages[$entity_type_id]);
    }

    foreach ($usages as $entity_type_id => $entity_type_usages) {
      foreach ($entity_type_usages as $entity_id => $entity_usages) {
        // Unset the source VID - we've already verified it's the latest
        // revision. It's possible the recorded source VID might no longer be
        // the latest one when we come to update the usage, so keeping it can
        // be actively misleading - just always update the latest revision in
        // the appropriate language.
        foreach ($entity_usages as $key => $batch_item) {
          unset($entity_usages[$key]['source_vid']);
        }
        $arguments = [
          $entity,
          $new_entity_type_id,
          $new_id,
          $entity_type_id,
          $entity_id,
          $entity_usages,
        ];
        $batch_builder->addOperation([static::class, 'updateReferences'], $arguments);
      }
    }

    $operation = [static::class, 'updateRevisionReferenceHostEntities'];
    $batch_builder->addOperation($operation);
    batch_set($batch_builder->toArray());
  }

  /**
   * Updates references within a single entity.
   *
   * It handles all translations for the given entity in one go.
   *
   * @param \Drupal\Core\Entity\EntityInterface $old_target
   *   The entity that should no longer be referenced.
   * @param string $new_target_entity_type
   *   The entity type of the entity that should be referenced instead of
   *   $old_target.
   * @param int|string $new_target_id
   *   The entity ID of the entity that should be referenced instead of
   *   $old_target.
   * @param string $entity_type_id
   *   The The entity type of the entity being updated.
   * @param int|string $entity_id
   *   The ID of the entity being updated.
   * @param string[][] $usages
   *   An array of associative arrays of usage information, keyed by:
   *   - source_langcode
   *   - method
   *   - field_name
   *   - count.
   * @param array &$context
   *   The batch API context array.
   */
  public static function updateReferences(EntityInterface $old_target, string $new_target_entity_type, $new_target_id, string $entity_type_id, $entity_id, array $usages, array &$context) {

    $storage = \Drupal::entityTypeManager()->getStorage($entity_type_id);
    $entity_type = \Drupal::entityTypeManager()->getDefinition($entity_type_id);
    if (!$storage instanceof RevisionableStorageInterface) {
      throw new \RuntimeException("Shouldn't be 'ere");
    }

    $storage = \Drupal::entityTypeManager()->getStorage($entity_type_id);

    // I'm not sure if this is the best way to support all possible multilingual
    // pending revision scenarios. We just load the latest revision, and then
    // update all the translations which need updating.
    if ($entity_type->isRevisionable()) {
      $vid = $storage->getLatestRevisionId($entity_id);
      $current_revision = $storage->loadRevision($vid);
    }
    else {
      $current_revision = $storage->load($entity_id);
    }

    if ($entity_type->isRevisionable()) {
      $new_revision = static::createRevision($current_revision);
    }
    else {
      $new_revision = $current_revision;
    }

    foreach ($usages as $reference) {
      $langcode = $reference['source_langcode'];

      if ($current_revision instanceof TranslatableInterface) {
        $current_translation = $current_revision->getTranslation($langcode);
        $new_translation = $new_revision->getTranslation($langcode);
      }
      else {
        $current_translation = $current_revision;
        $new_translation = $new_revision;
      }

      static::processRevisionReferences($current_translation, $context);
      static::updateUsage($old_target, $new_target_entity_type, $new_target_id, $new_translation, $reference, $context);
    }

    $new_revision->save();

    if (!isset($context['results']['update_count'])) {
      $context['results']['update_count'] = 0;
    }
    $context['results']['update_count']++;
  }

  /**
   * Tracks entities referenced via entity reference revision fields.
   *
   * If you update a field on an entity that's itself referenced via an ERR
   * field, then you also need to update the ERR field as otherwise it will
   * still refer to the earlier revision ID prior to the update. For example if
   * you update a link field on a paragraph, you need to make sure that any ERR
   * field referencing that paragraph is updated to reference the new paragraph
   * revision ID. This can be recursive: if you have nested paragraphs, you have
   * to keep updating references to the updated paragraph.
   *
   * For now this is dumb, and just checks if the entity is a paragraph. Other
   * entity types using ERR fields will not be updated. (But it's ok if there's
   * a non-ERR field referencing a paragraph, it will be tracked but won't cause
   * any problems.)
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity being updated.
   * @param array $context
   *   The batch context array.
   */
  protected static function processRevisionReferences(EntityInterface $entity, array &$context): void {
    // @todo More generally we want to consider any entity that's referenced by
    //   an ERR field. However I don't know if I can make the same assumptions
    //   about i18n with ERR fields in general that I can with paragraphs
    //   (because the paragraphs instructions are very clear).
    if ($entity->getEntityTypeId() === 'paragraph') {
      assert($entity instanceof TranslatableRevisionableInterface);
      // If we assume everything works like paragraphs wrt i18n then the
      // language is irrelevant. The ERR field is not translatable, and its
      // one value should be updated to the latest paragraph vid. That in turn
      // should have all the correct values for the various languages already.
      $context['results']['updated_err_entities'][] = [
        'id' => $entity->id(),
        'vid' => $entity->getRevisionId(),
        'entity_type' => $entity->getEntityTypeId(),
      ];
    }
  }

  /**
   * Updates references within a single usage on a single entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $old_target
   *   The entity that should no longer be referenced.
   * @param string $new_entity_type
   *   The entity type of the entity that should be referenced instead of
   *   $old_target.
   * @param int|string $new_id
   *   The entity ID of the entity that should be referenced instead of
   *   $old_target.
   * @param \Drupal\Core\Entity\FieldableEntityInterface $revision
   *   The revision being updated.
   * @param array $usage
   *   An associative array of usage information, keyed by
   *   - source_langcode
   *   - method
   *   - field_name
   *   - count.
   * @param array &$context
   *   The batch API context array.
   */
  public static function updateUsage(EntityInterface $old_target, string $new_entity_type, $new_id, FieldableEntityInterface $revision, array $usage, array &$context): void {
    if ($revision instanceof RevisionableInterface) {
      assert($revision->isLatestRevision());
    }
    assert($revision->language()->getId() === $usage['source_langcode']);

    // Use the Entity Usage tracker to help us identify which field deltas
    // contain references to the target entity. (That level of detail isn't
    // provided in the usage information.)
    $manager = \Drupal::service('plugin.manager.entity_usage.track');
    assert($manager instanceof PluginManagerInterface);
    $method = $usage['method'];
    $tracker = $manager->createInstance($method);

    $updater_manager = \Drupal::service('plugin.manager.entity_usage_updater');
    assert($updater_manager instanceof PluginManagerInterface);
    try {
      $updater = $updater_manager->createInstance($method);

      foreach ($revision->get($usage['field_name']) as $item) {
        // Use entity_usage to verify if the target entity is in the field.
        $targets = $tracker->getTargetEntities($item);
        $old_target_string = $old_target->getEntityTypeId() . '|' . $old_target->id();
        if ($targets && in_array($old_target_string, $targets)) {
          assert($updater instanceof EntityUsageUpdaterPluginInterface);
          $updater->update($old_target, $new_entity_type, $new_id, $item);
        }
      }

      if ($revision instanceof RevisionLogInterface) {
        $context = [
          '@old_entity_type' => $old_target->getEntityTypeId(),
          '@old_entity_id' => $old_target->id(),
          '@new_entity_type' => $new_entity_type,
          '@new_entity_id' => $new_id,
        ];
        $revision->setRevisionLogMessage(t("Automatically updated by the entity usage updater module to replace @old_entity_type @old_entity_id with @new_entity_type @new_entity_id", $context));
      }
    }
    catch (PluginException $e) {
      $context['results']['errors'][] = $e->getMessage();
    }
  }

  /**
   * Updates entity reference revision fields referencing updated entities.
   *
   * If you just update a paragraph that references an entity, then you won't
   * see any changes on the host entity; this is because entity reference
   * revision fields reference revisions, so the field is still explicitly
   * referencing the old revision. This function will find such "stale"
   * revisions and update the host entities to reference the latest revision.
   *
   * It can add to its own workload: for example if there are nested paragraphs
   * then when it updates the inner paragraph, it will update the context array
   * so that it will later update the outer one. The operation should not
   * complete until all such references have been dealt with. (So it's unknown
   * at the start of the operation how many iterations it will take.)
   *
   * @param array $context
   *   The batch API context array.
   */
  public static function updateRevisionReferenceHostEntities(array &$context): void {
    // Use sandbox for communication between iterations of the same operation.
    if (empty($context['sandbox']['host_entities']) && !empty($context['results']['updated_err_entities'])) {
      // It's possible this will cause a DB max packet error, depending on how
      // many entities are passed: it would be good to batch it as well. It's
      // also plausible that getHostEntities will exhaust memory, but it's not
      // loading entities, just querying for IDs and returning them, so less
      // likely to be an issue.
      $context['sandbox']['host_entities'] = static::getHostEntities(array_unique($context['results']['updated_err_entities']));
      $context['results']['updated_err_entities'] = [];
    }

    // Return if we've already handled all entities with updated ERR field
    // targets.
    if (empty($context['sandbox']['host_entities'])) {
      return;
    }

    $batch_size = Settings::get('entity_usage_updater_batch_size') ?? Settings::get('entity_update_batch_size', 50);

    $entity_type_id = key($context['sandbox']['host_entities']);
    $batch = array_slice($context['sandbox']['host_entities'][$entity_type_id], 0, $batch_size, TRUE);
    array_splice($context['sandbox']['host_entities'][$entity_type_id], 0, $batch_size);
    if (empty($context['sandbox']['host_entities'][$entity_type_id])) {
      unset($context['sandbox']['host_entities'][$entity_type_id]);
    }

    $ids = array_keys($batch);
    $repository = \Drupal::service('entity.repository');
    assert($repository instanceof EntityRepositoryInterface);

    foreach ($repository->getActiveMultiple($entity_type_id, $ids) as $entity) {
      assert($entity instanceof FieldableEntityInterface);
      // In case of nested ERR fields (eg. nested paragraphs) we still need to
      // track as we update.
      static::processRevisionReferences($entity, $context);

      $revision = static::createRevision($entity);
      foreach ($batch[$revision->id()] as $field_name) {
        // We're just updating the revision to the latest active.
        $field = $revision->{$field_name};
        assert($field instanceof FieldItemListInterface);
        $target_type = $field->getFieldDefinition()->getSetting('target_type');
        $field->target_revision_id = static::getActiveRevisionId($target_type, $field->target_id, $field->getLangcode());
      }

      // @todo A better log message might be nice, but it will involve storing
      //   more data than we currently do IIUC.
      if ($revision instanceof RevisionLogInterface) {
        $revision->setRevisionLogMessage(t("Automatically updated by the entity usage updater module."));
      }
      $revision->save();
      $context['results']['update_count']++;
    }

    $context['finished'] = 0;
  }

  /**
   * Queries for referencing host entities using ERR fields.
   *
   * It takes details of entities that have been updated and systematically goes
   * through all entity reference revision fields to see if any of them
   * reference updated entities, and returns the details of those that do. It
   * doesn't use the Entity Usage module because that might not be tracking
   * the host entities and it doesn't accept a revision ID to check for.
   *
   * @param array $entity_ids
   *   An array of associative arrays of now stale entity references, keyed by
   *   - id
   *   - vid
   *   - entity_type.
   *
   * @return array
   *   An array containing details of entities with ERR fields that now point to
   *   stale revisions and so need to be updated to the latest revision, eg.
   *   @code
   *   [
   *     'node' => [
   *       // 42 is *revision* ID
   *       42 => ['field_a', 'field_b'],
   *     ],
   *   ]
   *   @endcode
   */
  public static function getHostEntities(array $entity_ids): array {
    if (!$entity_ids) {
      return [];
    }
    $entity_field_manager = \Drupal::service('entity_field.manager');
    assert($entity_field_manager instanceof EntityFieldManagerInterface);
    $map = $entity_field_manager->getFieldMapByFieldType('entity_reference_revisions');
    $out = [];
    foreach ($map as $host_entity_type_id => $field_map) {
      $storage = \Drupal::entityTypeManager()->getStorage($host_entity_type_id);
      $field_storage_definitions = $entity_field_manager->getFieldStorageDefinitions($host_entity_type_id);
      foreach ($field_map as $field_name => $field_data) {

        // As ERR fields shouldn't be marked as translatable, we don't set the
        // language (which won't work with content moderation until #3088341
        // lands anyway).
        $query = $storage->getQuery()
          ->latestRevision()
          ->accessCheck(FALSE);

        $fields_conditions = $query->orConditionGroup();
        foreach ($entity_ids as ['id' => $entity_id, 'vid' => $revision_id, 'entity_type' => $entity_type_id]) {
          if ($field_storage_definitions[$field_name]->getSetting('target_type') === $entity_type_id) {
            $field_condition = $query->andConditionGroup();
            $field_condition->condition("$field_name.target_id", $entity_id);
            $field_condition->condition("$field_name.target_revision_id", $revision_id);
            $fields_conditions->condition($field_condition);
          }
        }
        $query->condition($fields_conditions);
        $results = $query->execute();
        // Ignore revision ID, just assume it's the active ID for the entity.
        // (In case the entity gets updated in the meantime by this script.)
        foreach ($results as $id) {
          $out[$host_entity_type_id][$id][] = $field_name;
        }
      }
    }
    return $out;
  }

  /**
   * Returns the active revision ID for a given entity.
   *
   * The active revision is appropriate for editing, though it might not be the
   * currently visible one. I'm not sure this plays nicely with all possible
   * forms of multilingual pending revisions.
   *
   * @param string $entity_type_id
   *   The entity type.
   * @param int|string $entity_id
   *   The entity ID.
   * @param string $langcode
   *   The language code.
   *
   * @return int
   *   The revision ID.
   *
   * @throws \RuntimeException
   *   If no revisions are found.
   */
  protected static function getActiveRevisionId(string $entity_type_id, $entity_id, string $langcode): int {
    $storage = \Drupal::entityTypeManager()->getStorage($entity_type_id);
    // Latest revision IDs are cached by ContentEntityStorageBase.
    assert($storage instanceof RevisionableStorageInterface);
    if ($storage instanceof TranslatableRevisionableStorageInterface) {
      $active_vid = $storage->getLatestTranslationAffectedRevisionId($entity_id, $langcode);
    }
    else {
      $active_vid = $storage->getLatestRevisionId($entity_id);
    }
    if ($active_vid === NULL) {
      throw new \RuntimeException("No revision found for $entity_type_id $entity_id $langcode");
    }
    return intval($active_vid);
  }

  /**
   * Creates a new revision from an entity/revision if possible.
   *
   * If a new revision is created it will be default if $entity was default.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to create a new revision of.
   *
   * @return \Drupal\Core\Entity\EntityInterface
   *   For revisionable entities a new revision; otherwise $entity.
   */
  protected static function createRevision(EntityInterface $entity): EntityInterface {
    if (!$entity instanceof RevisionableInterface) {
      return $entity;
    }

    $storage = \Drupal::entityTypeManager()
      ->getStorage($entity->getEntityTypeId());
    assert($storage instanceof RevisionableStorageInterface);
    return $storage->createRevision($entity, $entity->isDefaultRevision());
  }

  /**
   * Displays messages after finishing with a batch.
   *
   * It's fairly generic and used for different batches.
   *
   * @param bool $success
   *   Indicate that the batch API tasks were all completed successfully.
   * @param array $results
   *   An array of all the results that were updated in update_do_one().
   * @param array $operations
   *   A list of the operations that had not been completed by the batch API.
   */
  public static function finish($success, $results, $operations) {
    if (!$success) {
      \Drupal::messenger()->addError("The update didn't complete succesfully.");
    }

    $map = [
      'errors' => MessengerInterface::TYPE_ERROR,
      'warnings' => MessengerInterface::TYPE_WARNING,
      'messages' => MessengerInterface::TYPE_STATUS,
    ];
    foreach ($map as $key => $type) {
      if (!empty($results[$key])) {
        foreach ($results[$key] as $message) {
          \Drupal::messenger()->addMessage($message, $type);
        }
      }
    }

    if ($success && isset($results['update_count'])) {
      $message = \Drupal::translation()->formatPlural(
        $results['update_count'],
        "Updated one entity.",
        "Updated @count entities."
      );
      \Drupal::messenger()->addStatus($message);
    }
  }

}
