<?php declare(strict_types=1);

namespace Drupal\entity_usage_updater\Form;

use Drupal\Core\Entity\ContentEntityTypeInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\entity_usage_updater\EntityUsageUpdaterInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a form for updating entity references.
 */
class EntityUsageUpdateForm extends FormBase {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The entity usage updater.
   *
   * @var \Drupal\entity_usage_updater\EntityUsageUpdaterInterface
   */
  protected EntityUsageUpdaterInterface $entityUsageUpdater;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $form = parent::create($container);
    $form->entityTypeManager = $container->get('entity_type.manager');
    $form->entityUsageUpdater = $container->get('entity_usage_updater.updater');
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'entity_usage_updater_entity_usage_update';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $is_content_entity_type = fn(EntityTypeInterface $type): bool => $type instanceof ContentEntityTypeInterface;
    $types = array_filter($this->entityTypeManager->getDefinitions(), $is_content_entity_type);
    $options = [];
    foreach ($types as $type) {
      $options[$type->id()] = $type->getLabel();
    }
    asort($options, SORT_LOCALE_STRING);

    $form['entity_type_id'] = [
      '#title' => $this->t("Entity type"),
      '#type' => 'select',
      '#options' => $options,
      '#required' => TRUE,
      '#default_value' => isset($options['node']) ? 'node' : NULL,
    ];

    $form['old_entity_id'] = [
      '#title' => $this->t("Target entity ID"),
      '#description' => $this->t("Enter the ID of the entity you no longer want referenced."),
      '#type' => 'textfield',
      '#required' => TRUE,
      '#size' => 20,
    ];

    $form['new_entity_id'] = [
      '#title' => $this->t("Replacement entity ID"),
      '#description' => $this->t("Enter the ID of the entity you want to use instead."),
      '#type' => 'textfield',
      '#required' => TRUE,
      '#size' => 20,
    ];

    $form['actions'] = [
      '#type' => 'actions',
    ];
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t("Update"),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    $ids = [
      $form_state->getValue('old_entity_id'),
      $form_state->getValue('new_entity_id'),
    ];
    if ($ids[0] === $ids[1]) {
      $form_state->setError($form, $this->t("The target and replacement IDs must be different."));
    }
    $entity_type_id = $form_state->getValue('entity_type_id');
    $type = $this->entityTypeManager->getDefinition($entity_type_id);
    $query = $this->entityTypeManager->getStorage($entity_type_id)->getQuery();
    $results = $query->accessCheck(FALSE)
      ->condition($type->getKey('id'), $ids, 'IN')
      ->execute();
    if (count($results) !== 2) {
      $context = ['@type' => $type->getLabel()];
      foreach (['old_entity_id', 'new_entity_id'] as $name) {
        $id = $form_state->getValue($name);
        if (!in_array($id, $results, TRUE)) {
          $context['@id'] = $id;
          $form_state->setErrorByName($name, $this->t("@type with ID @id doesn't exist.", $context));
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $entity_type_id = $form_state->getValue('entity_type_id');
    $this->entityUsageUpdater->update([
      $entity_type_id => [
        $form_state->getValue('old_entity_id') => [
          $entity_type_id,
          $form_state->getValue('new_entity_id'),
        ],
      ],
    ]);
  }

}
