<?php declare(strict_types = 1);

namespace Drupal\entity_usage_updater\Form;

use Drupal\Component\Plugin\PluginManagerInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformState;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\Core\Render\Element;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a form to configure Entity Usage Updater.
 */
final class SettingsForm extends ConfigFormBase {

  /**
   * The entity usage updater plugin manager.
   *
   * @var \Drupal\Component\Plugin\PluginManagerInterface
   */
  protected PluginManagerInterface $pluginManager;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $form = parent::create($container);
    $form->pluginManager = $container->get('plugin.manager.entity_usage_updater');
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'entity_usage_updater_settings';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames(): array {
    return ['entity_usage_updater.settings'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {
    $form = parent::buildForm($form, $form_state);

    $form['#tree'] = TRUE;
    foreach ($this->pluginManager->getDefinitions() as $definition) {
      if (is_a($definition['class'], PluginFormInterface::class, TRUE)) {
        $id = $definition['id'];
        $instance = $this->createInstance($id);
        $form['plugin'][$id] = $instance->buildConfigurationForm([], $form_state);
      }
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state): void {
    parent::validateForm($form, $form_state);

    foreach (Element::children($form['plugin']) as $id) {
      $plugin_form = $form['plugin'][$id];
      $instance = $this->createInstance($id);
      $subform_state = SubformState::createForSubform($plugin_form, $form, $form_state);
      $instance->validateConfigurationForm($plugin_form, $subform_state);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    foreach (Element::children($form['plugin']) as $id) {
      $plugin_form = $form['plugin'][$id];
      $instance = $this->createInstance($id);
      $subform_state = SubformState::createForSubform($plugin_form, $form, $form_state);
      $instance->submitConfigurationForm($plugin_form, $subform_state);
    }

    $config = $this->config('entity_usage_updater.settings');
    foreach ($form_state->getValue('plugin') as $id => $configuration) {
      $config->set("plugins.$id", $configuration);
    }
    $config->save();

    parent::submitForm($form, $form_state);
  }

  /**
   * Creates an instance of the plugin with the stored configuration.
   *
   * @param string $id
   *   The plugin ID.
   *
   * @return \Drupal\Core\Plugin\PluginFormInterface
   *   The plugin.
   */
  protected function createInstance(string $id): PluginFormInterface {
    $config = $this->config('entity_usage_updater.settings');
    $configuration = $config->get('plugins.' . $id) ?? [];
    return $this->pluginManager->createInstance($id, $configuration);
  }

}
